import torch


def eye_permutation(n: int) -> torch.Tensor:
    """
    Returns toch.tensor([0, 1, 2, .., n-1])
    """
    return torch.arange(n)


def permute_list(input_list: list, perm: torch.Tensor) -> list:
    """
    Permutes the input list according to the given permutation.
    Parameters
    -------
    input_list :
        A list to permute.
    permutation :
        A list of indices representing the new order.
    Returns
    -------
        The permuted list.
    Example
    -------
    >>> permute_list(['a','b','c'], torch.tensor([2, 0, 1]))
    ['c', 'a', 'b']
    """
    return [input_list[i] for i in perm.tolist()]


def permute_tuple(input_tuple: tuple, perm: torch.Tensor) -> tuple:
    """
    Permutes the input tuple according to the given permutation.
    Parameters
    -------
    input_tuple :
        A tuple to permute.
    permutation :
        A tuple of indices representing the new order.
    Returns
    -------
        The permuted tuple.
    Example
    -------
    >>> permute_tuple(('a','b','c'), torch.tensor([2, 0, 1]))
    ('c', 'a', 'b')
    """
    lst_elem = list(input_tuple)
    return tuple(permute_list(lst_elem, perm))


def permute_string(input_str: str, perm: torch.Tensor) -> str:
    """
    Permutes the input string according to the given permutation.
    Parameters
    -------
    input_string :
        A string to permute.
    permutation :
        A list of indices representing the new order.
    Returns
    -------
        The permuted string.
    Example
    -------
    >>> permute_string("abc", torch.tensor([2, 0, 1]))
    'cab'
    """
    permuted = permute_list(list(input_str), perm)
    return "".join(permuted)


def inv_permutation(permutation: torch.Tensor) -> torch.Tensor:
    """
    inv_permutation(permutation) -> inverted_perm

    Inverts the input permutation list.

    Parameters
    -------
    permutation :
        A list of indices representing the order

    Returns
    -------
        permutation list inverse to the input list

    Example:
    -------
    >>> inv_permutation(torch.tensor([2, 0, 1]))
    tensor([1, 2, 0])
    """
    inv_perm = torch.empty_like(permutation)
    inv_perm[permutation] = torch.arange(len(permutation))
    return inv_perm


def permute_tensor(tensor: torch.Tensor, perm: torch.Tensor) -> torch.Tensor:
    """
    Permute a 1D or square 2D torch tensor using the given permutation indices.
    For 1D tensors, applies the permutation to the elements.
    For 2D square tensors, applies the same permutation to both rows and columns.

    Parameters
    ----------
    tensor : torch.Tensor
        A 1D or 2D square tensor to be permuted.
    perm : torch.Tensor
        A 1D tensor of indices specifying the permutation order.

    Returns
    -------
    torch.Tensor
        A new tensor with elements (1D) or rows and columns (2D) permuted according to `perm`.

    Raises
    ------
    ValueError
        If tensor is not 1D or square 2D.

    Examples
    --------
    >>> vector = torch.tensor([10, 20, 30])
    >>> perm = torch.tensor([2, 0, 1])
    >>> permute_tensor(vector, perm)
    tensor([30, 10, 20])

    >>> matrix = torch.tensor([
    ...     [1, 2, 3],
    ...     [4, 5, 6],
    ...     [7, 8, 9]])
    >>> perm = torch.tensor([1, 0, 2])
    >>> permute_tensor(matrix, perm)
    tensor([[5, 4, 6],
            [2, 1, 3],
            [8, 7, 9]])
    """
    if tensor.ndim == 1:
        return tensor[perm]
    elif tensor.ndim == 2 and tensor.shape[0] == tensor.shape[1]:
        return tensor[perm][:, perm]
    else:
        raise ValueError("Only 1D tensors or square 2D tensors are supported.")


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from typing import Any, ClassVar
from types import MethodType

import copy

from emu_base import DEVICE_COUNT
from emu_mps.solver import Solver
from emu_mps.custom_callback_implementations import (
    energy_mps_impl,
    energy_second_moment_mps_impl,
    energy_variance_mps_impl,
    correlation_matrix_mps_impl,
    qubit_occupation_mps_impl,
)
from pulser.backend import (
    Occupation,
    CorrelationMatrix,
    Energy,
    EnergySecondMoment,
    EnergyVariance,
    BitStrings,
    EmulationConfig,
)
import logging
import pathlib
import sys


class MPSConfig(EmulationConfig):
    """
    The configuration of the emu-mps MPSBackend. The kwargs passed to this class
    are passed on to the base class.
    See the API for that class for a list of available options.

    Args:
        dt: the timestep size that the solver uses. Note that observables are
            only calculated if the evaluation_times are divisible by dt.
        precision: up to what precision the state is truncated
        max_bond_dim: the maximum bond dimension that the state is allowed to have.
        max_krylov_dim:
            the size of the krylov subspace that the Lanczos algorithm maximally builds
        extra_krylov_tolerance:
            the Lanczos algorithm uses this*precision as the convergence tolerance
        num_gpus_to_use: during the simulation, distribute the state over this many GPUs
            0=all factors to cpu. As shown in the benchmarks, using multiple GPUs might
            alleviate memory pressure per GPU, but the runtime should be similar.
        optimize_qubit_ordering: Optimize the register ordering. Improves performance and
            accuracy, but disables certain features.
        interaction_cutoff: Set interaction coefficients below this value to `0`.
            Potentially improves runtime and memory consumption.
        log_level: How much to log. Set to `logging.WARN` to get rid of the timestep info.
        log_file: If specified, log to this file rather than stout.
        autosave_prefix: filename prefix for autosaving simulation state to file
        autosave_dt: minimum time interval in seconds between two autosaves.
            Saving the simulation state is only possible at specific times,
            therefore this interval is only a lower bound.
        solver: chooses the solver algorithm to run a sequence.
            Two options are currently available:
            ``TDVP``, which performs ordinary time evolution,
            and ``DMRG``, which adiabatically follows the ground state
            of a given adiabatic pulse.
        kwargs: arguments that are passed to the base class

    Examples:
        >>> num_gpus_to_use = 2 #use 2 gpus if available, otherwise 1 or cpu
        >>> dt = 1 #this will impact the runtime
        >>> precision = 1e-6 #smaller dt requires better precision, generally
        >>> MPSConfig(num_gpus_to_use=num_gpus_to_use, dt=dt, precision=precision,
        >>>     with_modulation=True) #the last arg is taken from the base class
    """

    # Whether to warn if unexpected kwargs are received
    _enforce_expected_kwargs: ClassVar[bool] = True

    def __init__(
        self,
        *,
        dt: int = 10,
        precision: float = 1e-5,
        max_bond_dim: int = 1024,
        max_krylov_dim: int = 100,
        extra_krylov_tolerance: float = 1e-3,
        num_gpus_to_use: int = DEVICE_COUNT,
        optimize_qubit_ordering: bool = False,
        interaction_cutoff: float = 0.0,
        log_level: int = logging.INFO,
        log_file: pathlib.Path | None = None,
        autosave_prefix: str = "emu_mps_save_",
        autosave_dt: int = 600,  # 10 minutes
        solver: Solver = Solver.TDVP,
        **kwargs: Any,
    ):
        kwargs.setdefault("observables", [BitStrings(evaluation_times=[1.0])])
        super().__init__(
            dt=dt,
            precision=precision,
            max_bond_dim=max_bond_dim,
            max_krylov_dim=max_krylov_dim,
            extra_krylov_tolerance=extra_krylov_tolerance,
            num_gpus_to_use=num_gpus_to_use,
            optimize_qubit_ordering=optimize_qubit_ordering,
            interaction_cutoff=interaction_cutoff,
            log_level=log_level,
            log_file=log_file,
            autosave_prefix=autosave_prefix,
            autosave_dt=autosave_dt,
            solver=solver,
            **kwargs,
        )
        if self.optimize_qubit_ordering:
            self.check_permutable_observables()

        MIN_AUTOSAVE_DT = 10
        assert (
            self.autosave_dt > MIN_AUTOSAVE_DT
        ), f"autosave_dt must be larger than {MIN_AUTOSAVE_DT} seconds"

        self.monkeypatch_observables()

        self.logger = logging.getLogger("global_logger")
        if log_file is None:
            logging.basicConfig(
                level=log_level, format="%(message)s", stream=sys.stdout, force=True
            )  # default to stream = sys.stderr
        else:
            logging.basicConfig(
                level=log_level,
                format="%(message)s",
                filename=str(log_file),
                filemode="w",
                force=True,
            )
        if (self.noise_model.runs != 1 and self.noise_model.runs is not None) or (
            self.noise_model.samples_per_run != 1
            and self.noise_model.samples_per_run is not None
        ):
            self.logger.warning(
                "Warning: The runs and samples_per_run values of the NoiseModel are ignored!"
            )

    def _expected_kwargs(self) -> set[str]:
        return super()._expected_kwargs() | {
            "dt",
            "precision",
            "max_bond_dim",
            "max_krylov_dim",
            "extra_krylov_tolerance",
            "num_gpus_to_use",
            "optimize_qubit_ordering",
            "interaction_cutoff",
            "log_level",
            "log_file",
            "autosave_prefix",
            "autosave_dt",
            "solver",
        }

    def monkeypatch_observables(self) -> None:
        obs_list = []
        for _, obs in enumerate(self.observables):  # monkey patch
            obs_copy = copy.deepcopy(obs)
            if isinstance(obs, Occupation):
                obs_copy.apply = MethodType(  # type: ignore[method-assign]
                    qubit_occupation_mps_impl, obs_copy
                )
            elif isinstance(obs, EnergyVariance):
                obs_copy.apply = MethodType(  # type: ignore[method-assign]
                    energy_variance_mps_impl, obs_copy
                )
            elif isinstance(obs, EnergySecondMoment):
                obs_copy.apply = MethodType(  # type: ignore[method-assign]
                    energy_second_moment_mps_impl, obs_copy
                )
            elif isinstance(obs, CorrelationMatrix):
                obs_copy.apply = MethodType(  # type: ignore[method-assign]
                    correlation_matrix_mps_impl, obs_copy
                )
            elif isinstance(obs, Energy):
                obs_copy.apply = MethodType(  # type: ignore[method-assign]
                    energy_mps_impl, obs_copy
                )
            obs_list.append(obs_copy)
        self.observables = tuple(obs_list)

    def init_logging(self) -> None:
        if self.log_file is None:
            logging.basicConfig(
                level=self.log_level, format="%(message)s", stream=sys.stdout, force=True
            )  # default to stream = sys.stderr
        else:
            logging.basicConfig(
                level=self.log_level,
                format="%(message)s",
                filename=str(self.log_file),
                filemode="w",
                force=True,
            )

    def check_permutable_observables(self) -> None:
        allowed_permutable_obs = set(
            [
                "bitstrings",
                "occupation",
                "correlation_matrix",
                "statistics",
                "energy",
                "energy_variance",
                "energy_second_moment",
            ]
        )

        actual_obs = set([obs._base_tag for obs in self.observables])
        not_allowed = actual_obs.difference(allowed_permutable_obs)
        if not_allowed:
            raise ValueError(
                f"emu-mps allows only {allowed_permutable_obs} observables with"
                " `optimize_qubit_ordering = True`."
                f" you provided unsupported {not_allowed}"
                " To use other observables, please set"
                " `optimize_qubit_ordering = False` in `MPSConfig()`."
            )



from .optimiser import minimize_bandwidth
from .permutations import (
    permute_tensor,
    inv_permutation,
    permute_string,
    eye_permutation,
    permute_list,
    permute_tuple,
)


__all__ = [
    "minimize_bandwidth",
    "eye_permutation",
    "permute_string",
    "permute_tensor",
    "inv_permutation",
    "permute_list",
    "permute_tuple",
]



import torch

from pulser.backend.default_observables import (
    CorrelationMatrix,
    EnergySecondMoment,
    EnergyVariance,
    Occupation,
    Energy,
)
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from emu_mps.mps_config import MPSConfig
    from emu_mps.mps import MPS
    from emu_mps.mpo import MPO


def qubit_occupation_mps_impl(
    self: Occupation,
    *,
    config: "MPSConfig",
    state: "MPS",
    hamiltonian: "MPO",
) -> torch.Tensor:
    """
    Custom implementation of the occupation ❬ψ|nᵢ|ψ❭ for the EMU-MPS.
    """
    op = torch.tensor(
        [[[0.0, 0.0], [0.0, 1.0]]], dtype=torch.complex128, device=state.factors[0].device
    )
    return state.expect_batch(op).real.view(-1).cpu()


def correlation_matrix_mps_impl(
    self: CorrelationMatrix,
    *,
    config: "MPSConfig",
    state: "MPS",
    hamiltonian: "MPO",
) -> torch.Tensor:
    """
    Custom implementation of the density-density correlation ❬ψ|nᵢnⱼ|ψ❭ for the EMU-MPS.

    TODO: extend to arbitrary two-point correlation ❬ψ|AᵢBⱼ|ψ❭
    """
    return state.get_correlation_matrix().cpu()


def energy_variance_mps_impl(
    self: EnergyVariance,
    *,
    config: "MPSConfig",
    state: "MPS",
    hamiltonian: "MPO",
) -> torch.Tensor:
    """
    Custom implementation of the energy variance ❬ψ|H²|ψ❭-❬ψ|H|ψ❭² for the EMU-MPS.
    """
    h_squared = hamiltonian @ hamiltonian
    h_2 = h_squared.expect(state).cpu()
    h = hamiltonian.expect(state).cpu()
    en_var = h_2 - h**2
    return en_var.real  # type: ignore[no-any-return]


def energy_second_moment_mps_impl(
    self: EnergySecondMoment,
    *,
    config: "MPSConfig",
    state: "MPS",
    hamiltonian: "MPO",
) -> torch.Tensor:
    """
    Custom implementation of the second moment of energy ❬ψ|H²|ψ❭
    for the EMU-MPS.
    """
    h_square = hamiltonian @ hamiltonian
    h_2 = h_square.expect(state).cpu()
    assert torch.allclose(h_2.imag, torch.zeros_like(h_2.imag), atol=1e-4)
    return h_2.real


def energy_mps_impl(
    self: Energy,
    *,
    config: "MPSConfig",
    state: "MPS",
    hamiltonian: "MPO",
) -> torch.Tensor:
    """
    Custom implementation of the second moment of energy ❬ψ|H²|ψ❭
    for the EMU-MPS.
    """
    h = hamiltonian.expect(state)
    assert torch.allclose(h.imag, torch.zeros_like(h.imag), atol=1e-4)
    return h.real



from __future__ import annotations

import math
from collections import Counter
from typing import List, Optional, Sequence, TypeVar, Mapping

import torch

from pulser.backend.state import State, Eigenstate
from emu_base import DEVICE_COUNT, apply_measurement_errors
from emu_mps import MPSConfig
from emu_mps.algebra import add_factors, scale_factors
from emu_mps.utils import (
    assign_devices,
    truncate_impl,
    tensor_trace,
    n_operator,
)

ArgScalarType = TypeVar("ArgScalarType")


class MPS(State[complex, torch.Tensor]):
    """
    Matrix Product State, aka tensor train.

    Each tensor has 3 dimensions ordered as such: (left bond, site, right bond).

    Only qubits are supported.
    """

    def __init__(
        self,
        factors: List[torch.Tensor],
        /,
        *,
        orthogonality_center: Optional[int] = None,
        config: Optional[MPSConfig] = None,
        num_gpus_to_use: Optional[int] = DEVICE_COUNT,
        eigenstates: Sequence[Eigenstate] = ("r", "g"),
    ):
        """
        This constructor creates a MPS directly from a list of tensors. It is for internal use only.

        Args:
            factors: the tensors for each site
                WARNING: for efficiency in a lot of use cases, this list of tensors
                IS NOT DEEP-COPIED. Therefore, the new MPS object is not necessarily
                the exclusive owner of the list and its tensors. As a consequence,
                beware of potential external modifications affecting the list or the tensors.
                You are responsible for deciding whether to pass its own exclusive copy
                of the data to this constructor, or some shared objects.
            orthogonality_center: the orthogonality center of the MPS, or None (in which case
                it will be orthogonalized when needed)
            config: the emu-mps config object passed to the run method
            num_gpus_to_use: distribute the factors over this many GPUs
                0=all factors to cpu, None=keep the existing device assignment.
        """
        super().__init__(eigenstates=eigenstates)
        self.config = config if config is not None else MPSConfig()
        assert all(
            factors[i - 1].shape[2] == factors[i].shape[0] for i in range(1, len(factors))
        ), "The dimensions of consecutive tensors should match"
        assert (
            factors[0].shape[0] == 1 and factors[-1].shape[2] == 1
        ), "The dimension of the left (right) link of the first (last) tensor should be 1"

        self.factors = factors
        self.num_sites = len(factors)
        assert self.num_sites > 1  # otherwise, do state vector

        assert (orthogonality_center is None) or (
            0 <= orthogonality_center < self.num_sites
        ), "Invalid orthogonality center provided"
        self.orthogonality_center = orthogonality_center

        if num_gpus_to_use is not None:
            assign_devices(self.factors, min(DEVICE_COUNT, num_gpus_to_use))

    @property
    def n_qudits(self) -> int:
        """The number of qudits in the state."""
        return self.num_sites

    @classmethod
    def make(
        cls,
        num_sites: int,
        config: Optional[MPSConfig] = None,
        num_gpus_to_use: int = DEVICE_COUNT,
        eigenstates: Sequence[Eigenstate] = ["0", "1"],
    ) -> MPS:
        """
        Returns a MPS in ground state |000..0>.

        Args:
            num_sites: the number of qubits
            config: the MPSConfig
            num_gpus_to_use: distribute the factors over this many GPUs
                0=all factors to cpu
        """
        config = config if config is not None else MPSConfig()

        if num_sites <= 1:
            raise ValueError("For 1 qubit states, do state vector")

        return cls(
            [
                torch.tensor([[[1.0], [0.0]]], dtype=torch.complex128)
                for _ in range(num_sites)
            ],
            config=config,
            num_gpus_to_use=num_gpus_to_use,
            orthogonality_center=0,  # Arbitrary: every qubit is an orthogonality center.
            eigenstates=eigenstates,
        )

    def __repr__(self) -> str:
        result = "["
        for fac in self.factors:
            result += repr(fac)
            result += ", "
        result += "]"
        return result

    def orthogonalize(self, desired_orthogonality_center: int = 0) -> int:
        """
        Orthogonalize the state on the given orthogonality center.

        Returns the new orthogonality center index as an integer,
        this is convenient for type-checking purposes.
        """
        assert (
            0 <= desired_orthogonality_center < self.num_sites
        ), f"Cannot move orthogonality center to nonexistent qubit #{desired_orthogonality_center}"

        lr_swipe_start = (
            self.orthogonality_center if self.orthogonality_center is not None else 0
        )

        for i in range(lr_swipe_start, desired_orthogonality_center):
            q, r = torch.linalg.qr(self.factors[i].view(-1, self.factors[i].shape[2]))
            self.factors[i] = q.view(self.factors[i].shape[0], 2, -1)
            self.factors[i + 1] = torch.tensordot(
                r.to(self.factors[i + 1].device), self.factors[i + 1], dims=1
            )

        rl_swipe_start = (
            self.orthogonality_center
            if self.orthogonality_center is not None
            else (self.num_sites - 1)
        )

        for i in range(rl_swipe_start, desired_orthogonality_center, -1):
            q, r = torch.linalg.qr(
                self.factors[i].contiguous().view(self.factors[i].shape[0], -1).mT,
            )
            self.factors[i] = q.mT.view(-1, 2, self.factors[i].shape[2])
            self.factors[i - 1] = torch.tensordot(
                self.factors[i - 1], r.to(self.factors[i - 1].device), ([2], [1])
            )

        self.orthogonality_center = desired_orthogonality_center

        return desired_orthogonality_center

    def truncate(self) -> None:
        """
        SVD based truncation of the state. Puts the orthogonality center at the first qubit.
        Calls orthogonalize on the last qubit, and then sweeps a series of SVDs right-left.
        Uses self.config for determining accuracy.
        An in-place operation.
        """
        self.orthogonalize(self.num_sites - 1)
        truncate_impl(self.factors, config=self.config)
        self.orthogonality_center = 0

    def get_max_bond_dim(self) -> int:
        """
        Return the max bond dimension of this MPS.

        Returns:
            the largest bond dimension in the state
        """
        return max((x.shape[2] for x in self.factors), default=0)

    def sample(
        self,
        *,
        num_shots: int,
        one_state: Eigenstate | None = None,
        p_false_pos: float = 0.0,
        p_false_neg: float = 0.0,
    ) -> Counter[str]:
        """
        Samples bitstrings, taking into account the specified error rates.

        Args:
            num_shots: how many bitstrings to sample
            p_false_pos: the rate at which a 0 is read as a 1
            p_false_neg: the rate at which a 1 is read as a 0

        Returns:
            the measured bitstrings, by count
        """
        assert one_state in {None, "r", "1"}
        self.orthogonalize(0)

        rnd_matrix = torch.rand(num_shots, self.num_sites).to(self.factors[0].device)

        bitstrings: Counter[str] = Counter()

        # Shots are performed in batches.
        # Larger max_batch_size is faster but uses more memory.
        max_batch_size = 32

        shots_done = 0
        while shots_done < num_shots:
            batch_size = min(max_batch_size, num_shots - shots_done)
            batched_accumulator = torch.ones(
                batch_size, 1, dtype=torch.complex128, device=self.factors[0].device
            )

            batch_outcomes = torch.empty(batch_size, self.num_sites, dtype=torch.bool)

            for qubit, factor in enumerate(self.factors):
                batched_accumulator = torch.tensordot(
                    batched_accumulator.to(factor.device), factor, dims=1
                )

                # Probability of measuring qubit == 0 for each shot in the batch
                probas = (
                    torch.linalg.vector_norm(batched_accumulator[:, 0, :], dim=1) ** 2
                )

                outcomes = (
                    rnd_matrix[shots_done : shots_done + batch_size, qubit].to(
                        factor.device
                    )
                    > probas
                )
                batch_outcomes[:, qubit] = outcomes

                # Batch collapse qubit
                tmp = torch.stack((~outcomes, outcomes), dim=1).to(dtype=torch.complex128)

                batched_accumulator = (
                    torch.tensordot(batched_accumulator, tmp, dims=([1], [1]))
                    .diagonal(dim1=0, dim2=2)
                    .transpose(1, 0)
                )
                batched_accumulator /= torch.sqrt(
                    (~outcomes) * probas + outcomes * (1 - probas)
                ).unsqueeze(1)

            shots_done += batch_size

            for outcome in batch_outcomes:
                bitstrings.update(["".join("0" if x == 0 else "1" for x in outcome)])

        if p_false_neg > 0 or p_false_pos > 0:
            bitstrings = apply_measurement_errors(
                bitstrings,
                p_false_pos=p_false_pos,
                p_false_neg=p_false_neg,
            )
        return bitstrings

    def norm(self) -> torch.Tensor:
        """Computes the norm of the MPS."""
        orthogonality_center = (
            self.orthogonality_center
            if self.orthogonality_center is not None
            else self.orthogonalize(0)
        )
        # the torch.norm function is not properly typed.
        return self.factors[orthogonality_center].norm().cpu()  # type: ignore[no-any-return]

    def inner(self, other: State) -> torch.Tensor:
        """
        Compute the inner product between this state and other.
        Note that self is the left state in the inner product,
        so this function is linear in other, and anti-linear in self

        Args:
            other: the other state

        Returns:
            inner product
        """
        assert isinstance(other, MPS), "Other state also needs to be an MPS"
        assert (
            self.num_sites == other.num_sites
        ), "States do not have the same number of sites"

        acc = torch.ones(1, 1, dtype=self.factors[0].dtype, device=self.factors[0].device)

        for i in range(self.num_sites):
            acc = acc.to(self.factors[i].device)
            acc = torch.tensordot(acc, other.factors[i].to(acc.device), dims=1)
            acc = torch.tensordot(self.factors[i].conj(), acc, dims=([0, 1], [0, 1]))

        return acc.view(1)[0].cpu()

    def overlap(self, other: State, /) -> torch.Tensor:
        """
        Compute the overlap of this state and other. This is defined as
        $|\\langle self | other \\rangle |^2$
        """
        return torch.abs(self.inner(other)) ** 2  # type: ignore[no-any-return]

    def entanglement_entropy(self, mps_site: int) -> torch.Tensor:
        """
        Returns
        the Von Neumann entanglement entropy of the state `mps` at the bond between sites b and b+1
        S = -Σᵢsᵢ² log(sᵢ²)),
        where sᵢ are the singular values at the chosen bond.
        """
        self.orthogonalize(mps_site)

        # perform svd on reshaped matrix at site b
        matrix = self.factors[mps_site].flatten(end_dim=1)
        s = torch.linalg.svdvals(matrix)

        s_e = torch.Tensor(torch.special.entr(s**2))
        s_e = torch.sum(s_e)

        self.orthogonalize(0)
        return s_e.cpu()

    def get_memory_footprint(self) -> float:
        """
        Returns the number of MBs of memory occupied to store the state

        Returns:
            the memory in MBs
        """
        return (  # type: ignore[no-any-return]
            sum(factor.element_size() * factor.numel() for factor in self.factors) * 1e-6
        )

    def __add__(self, other: State) -> MPS:
        """
        Returns the sum of two MPSs, computed with a direct algorithm.
        The resulting MPS is orthogonalized on the first site and truncated
        up to `self.config.precision`.

        Args:
            other: the other state

        Returns:
            the summed state
        """
        assert isinstance(other, MPS), "Other state also needs to be an MPS"
        assert (
            self.eigenstates == other.eigenstates
        ), f"`Other` state has basis {other.eigenstates} != {self.eigenstates}"
        new_tt = add_factors(self.factors, other.factors)
        result = MPS(
            new_tt,
            config=self.config,
            num_gpus_to_use=None,
            orthogonality_center=None,  # Orthogonality is lost.
            eigenstates=self.eigenstates,
        )
        result.truncate()
        return result

    def __rmul__(self, scalar: complex) -> MPS:
        """
        Multiply an MPS by a scalar.

        Args:
            scalar: the scale factor

        Returns:
            the scaled MPS
        """
        which = (
            self.orthogonality_center
            if self.orthogonality_center is not None
            else 0  # No need to orthogonalize for scaling.
        )
        factors = scale_factors(self.factors, scalar, which=which)
        return MPS(
            factors,
            config=self.config,
            num_gpus_to_use=None,
            orthogonality_center=self.orthogonality_center,
            eigenstates=self.eigenstates,
        )

    def __imul__(self, scalar: complex) -> MPS:
        return self.__rmul__(scalar)

    @classmethod
    def _from_state_amplitudes(
        cls,
        *,
        eigenstates: Sequence[Eigenstate],
        n_qudits: int,
        amplitudes: Mapping[str, complex],
    ) -> tuple[MPS, Mapping[str, complex]]:
        """
        See the base class.

        Args:
            basis: A tuple containing the basis states (e.g., ('r', 'g')).
            nqubits: the number of qubits.
            strings: A dictionary mapping state strings to complex or floats amplitudes.

        Returns:
            The resulting MPS representation of the state.s
        """
        basis = set(eigenstates)
        if basis == {"r", "g"}:
            one = "r"
        elif basis == {"0", "1"}:
            one = "1"
        else:
            raise ValueError("Unsupported basis provided")

        basis_0 = torch.tensor([[[1.0], [0.0]]], dtype=torch.complex128)  # ground state
        basis_1 = torch.tensor([[[0.0], [1.0]]], dtype=torch.complex128)  # excited state

        accum_mps = MPS(
            [torch.zeros((1, 2, 1), dtype=torch.complex128)] * n_qudits,
            orthogonality_center=0,
            eigenstates=eigenstates,
        )

        for state, amplitude in amplitudes.items():
            factors = [basis_1 if ch == one else basis_0 for ch in state]
            accum_mps += amplitude * MPS(factors, eigenstates=eigenstates)
        norm = accum_mps.norm()
        if not math.isclose(1.0, norm, rel_tol=1e-5, abs_tol=0.0):
            print("\nThe state is not normalized, normalizing it for you.")
            accum_mps *= 1 / norm

        return accum_mps, amplitudes

    def expect_batch(self, single_qubit_operators: torch.Tensor) -> torch.Tensor:
        """
        Computes expectation values for each qubit and each single qubit operator in
        the batched input tensor.

        Returns a tensor T such that T[q, i] is the expectation value for qubit #q
        and operator single_qubit_operators[i].
        """
        orthogonality_center = (
            self.orthogonality_center
            if self.orthogonality_center is not None
            else self.orthogonalize(0)
        )

        result = torch.zeros(
            self.num_sites, single_qubit_operators.shape[0], dtype=torch.complex128
        )

        center_factor = self.factors[orthogonality_center]
        for qubit_index in range(orthogonality_center, self.num_sites):
            temp = torch.tensordot(center_factor.conj(), center_factor, ([0, 2], [0, 2]))

            result[qubit_index] = torch.tensordot(
                single_qubit_operators.to(temp.device), temp, dims=2
            )

            if qubit_index < self.num_sites - 1:
                _, r = torch.linalg.qr(center_factor.view(-1, center_factor.shape[2]))
                center_factor = torch.tensordot(
                    r, self.factors[qubit_index + 1].to(r.device), dims=1
                )

        center_factor = self.factors[orthogonality_center]
        for qubit_index in range(orthogonality_center - 1, -1, -1):
            _, r = torch.linalg.qr(
                center_factor.view(center_factor.shape[0], -1).mT,
            )
            center_factor = torch.tensordot(
                self.factors[qubit_index],
                r.to(self.factors[qubit_index].device),
                ([2], [1]),
            )

            temp = torch.tensordot(center_factor.conj(), center_factor, ([0, 2], [0, 2]))

            result[qubit_index] = torch.tensordot(
                single_qubit_operators.to(temp.device), temp, dims=2
            )

        return result

    def apply(self, qubit_index: int, single_qubit_operator: torch.Tensor) -> None:
        """
        Apply given single qubit operator to qubit qubit_index, leaving the MPS
        orthogonalized on that qubit.
        """
        self.orthogonalize(qubit_index)

        self.factors[qubit_index] = (
            single_qubit_operator.to(self.factors[qubit_index].device)
            @ self.factors[qubit_index]
        )

    def get_correlation_matrix(
        self, *, operator: torch.Tensor = n_operator
    ) -> torch.Tensor:
        """
        Efficiently compute the symmetric correlation matrix
            C_ij = <self|operator_i operator_j|self>
        in basis ("r", "g").

        Args:
            operator: a 2x2 Torch tensor to use

        Returns:
            the corresponding correlation matrix
        """
        assert operator.shape == (2, 2)

        result = torch.zeros(self.num_sites, self.num_sites, dtype=torch.complex128)

        for left in range(0, self.num_sites):
            self.orthogonalize(left)
            accumulator = torch.tensordot(
                self.factors[left],
                operator.to(self.factors[left].device),
                dims=([1], [0]),
            )
            accumulator = torch.tensordot(
                accumulator, self.factors[left].conj(), dims=([0, 2], [0, 1])
            )
            result[left, left] = accumulator.trace().item().real
            for right in range(left + 1, self.num_sites):
                partial = torch.tensordot(
                    accumulator.to(self.factors[right].device),
                    self.factors[right],
                    dims=([0], [0]),
                )
                partial = torch.tensordot(
                    partial, self.factors[right].conj(), dims=([0], [0])
                )

                result[left, right] = (
                    torch.tensordot(
                        partial, operator.to(partial.device), dims=([0, 2], [0, 1])
                    )
                    .trace()
                    .item()
                    .real
                )
                result[right, left] = result[left, right]
                accumulator = tensor_trace(partial, 0, 2)

        return result


def inner(left: MPS, right: MPS) -> torch.Tensor:
    """
    Wrapper around MPS.inner.

    Args:
        left: the anti-linear argument
        right: the linear argument

    Returns:
        the inner product
    """
    return left.inner(right)



from pulser.backend.state import State
from pulser.backend.observable import Observable
from emu_mps.mps import MPS
from typing import Sequence, Any
import torch


class EntanglementEntropy(Observable):
    """Entanglement Entropy subclass used only in emu_mps"""

    def __init__(
        self,
        mps_site: int,
        *,
        evaluation_times: Sequence[float] | None = None,
        tag_suffix: str | None = None,
    ):
        super().__init__(evaluation_times=evaluation_times, tag_suffix=tag_suffix)
        self.mps_site = mps_site

    @property
    def _base_tag(self) -> str:
        return "entanglement_entropy"

    def _to_abstract_repr(self) -> dict[str, Any]:
        repr = super()._to_abstract_repr()
        repr["mps_site"] = self.mps_site
        return repr

    def apply(self, *, state: State, **kwargs: Any) -> torch.Tensor:
        if not isinstance(state, MPS):
            raise NotImplementedError(
                "Entanglement entropy observable is only available for emu_mps emulator."
            )
        if not (0 <= self.mps_site <= len(state.factors) - 2):
            raise ValueError(
                f"Invalid bond index {self.mps_site}. "
                f"Expected value in range 0 <= bond_index <= {len(state.factors)-2}."
            )
        return state.entanglement_entropy(self.mps_site)
